/*	Public domain	*/

#ifndef _FREESG_MAP_MAP_H_
#define _FREESG_MAP_MAP_H_

#include <freesg/rg/tileset.h>

#define MAPTILESZ 16		/* Default tile size in pixels */

#define MAP_TILESZ_MAX		16384
#define MAP_WIDTH_MAX		32767
#define MAP_HEIGHT_MAX		32767
#define MAP_LAYERS_MAX		256
#define MAP_CAMERAS_MAX		256
#define MAP_LAYER_NAME_MAX	128
#define MAP_CAMERA_NAME_MAX	128
#define MAP_NODE_ITEMS_MAX	32767
#define MAP_ITEM_MAXMASKS	16384

#include <freesg/map/nodemask.h>

#include <freesg/map/begin.h>

enum map_item_type {
	MAP_ITEM_TILE,		/* Reference to a tile */
	MAP_ITEM_ANIM,		/* Reference to an animation */
	MAP_ITEM_WARP		/* Reference to another location */
};

typedef struct map_item {
	enum map_item_type type;	/* Type of element */
	
	Uint flags;
#define MAP_ITEM_BLOCK		0x001	/* Tile block */
#define MAP_ITEM_CLIMBABLE	0x002	/* Surface is climbable */
#define MAP_ITEM_SLIPPERY	0x004	/* Surface is slippery */
#define MAP_ITEM_JUMPABLE	0x008	/* Element is jumpable */
#define MAP_ITEM_NOSAVE		0x100	/* Non persistent */
#define MAP_ITEM_MOUSEOVER	0x200	/* Mouse overlap (for editor) */
#define MAP_ITEM_SELECTED	0x400	/* Selection (for editor) */

	Sint8 friction;			/* Coefficient of friction */
	Uint8 layer;			/* Associated layer */
	void *p;			/* User pointer (non-persistent) */

	struct {
		Sint16 xcenter, ycenter;	/* Centering offsets */
		Sint16 xmotion, ymotion;	/* Motion offsets */
		Sint16 xorigin, yorigin;	/* Origin point */
		AG_Rect rs;			/* Source rectangle */
	} r_gfx;
	union {
		struct {
			RG_Tileset *obj;	/* Tileset object */
			Uint id;		/* Tile ID */
		} tile;
		struct {
			RG_Tileset *obj;	/* Tileset object */
			Uint id;		/* Animation ID */
			Uint *curframe;		/* Current frame number */
		} anim;
		struct {
			char *map;		/* Map identifier */
			int x, y;		/* Origin override */
			Uint8 dir;		/* Default direction */
		} warp;
	} nref;
	RG_TransformChain transforms;		/* Graphical transformations */
	MAP_NodeMaskQ masks;			/* Collision detection masks */
	AG_TAILQ_ENTRY(map_item) nrefs;		/* Node's reference stack */
#define r_tile		nref.tile
#define r_anim		nref.anim
#define r_warp		nref.warp
} MAP_Item;

AG_TAILQ_HEAD(map_itemq, map_item);

typedef struct map_node {
	struct map_itemq nrefs;		/* Items on this node */
} MAP_Node;

typedef struct map_layer {
	char name[MAP_LAYER_NAME_MAX];
	int visible;				/* Show/hide flag */
	Sint16 xinc, yinc;			/* Rendering direction */
	Uint8 alpha;				/* Transparency value */
} MAP_Layer;

enum map_camera_alignment {
	AG_MAP_UPPER_LEFT,
	AG_MAP_MIDDLE_LEFT,
	AG_MAP_LOWER_LEFT,
	AG_MAP_UPPER_RIGHT,
	AG_MAP_MIDDLE_RIGHT,
	AG_MAP_LOWER_RIGHT,
	AG_MAP_CENTER,
	AG_MAP_LOWER_CENTER,
	AG_MAP_UPPER_CENTER
};

typedef struct map_camera {
	char name[MAP_CAMERA_NAME_MAX];
	int flags;
	int x, y;				/* Position of camera */
	enum map_camera_alignment alignment;	/* View alignment */
	Uint zoom;				/* Zoom (%) */
	int tilesz;				/* Tile size */
	int pixsz;				/* Scaled pixel size */
} MAP_Camera;

enum map_mod_type {
	AG_MAPMOD_NODECHG,
	AG_MAPMOD_LAYERADD,
	AG_MAPMOD_LAYERDEL
};

typedef struct map_mod {
	enum map_mod_type type;
	union {
		struct {
			MAP_Node node;
			int x, y;
		} nodechg;
		struct {
			int nlayer;
		} layermod;
	} data;
#define mm_nodechg data.nodechg
#define mm_layeradd data.layermod
#define mm_layerdel data.layermod
} MAP_Mod;

typedef struct map_mod_blk {
	MAP_Mod *mods;
	Uint nmods;
	Uchar cancel;
} MAP_ModBlk;

struct map_actor;

typedef struct map {
	struct ag_object obj;

	AG_Mutex lock;
	Uint flags;
#define AG_MAP_SAVE_CAM0POS	0x01	/* Save the camera 0 position */
#define AG_MAP_SAVE_CAM0ZOOM	0x02	/* Save the camera 0 zoom factor */
#define AG_MAP_SAVED_FLAGS	(AG_MAP_SAVE_CAM0POS|AG_MAP_SAVE_CAM0ZOOM)

	Uint mapw, maph;		/* Map geometry */
	int cur_layer;			/* Layer being edited */
	struct {
		int x, y;		/* Origin coordinates */
		int layer;		/* Default tile layer */
	} origin;
	MAP_Node **map;			/* Arrays of nodes */
	int redraw;			/* Redraw (for tile-based mode) */
	MAP_Layer *layers;		/* Layer descriptions */
	Uint nlayers;
	MAP_Camera *cameras;		/* Views */
	Uint ncameras;
	MAP_ModBlk *blks;		/* Saved modifications */
	Uint nblks;
	Uint curblk;
	Uint nmods;
	AG_TAILQ_HEAD_(map_actor) actors;	/* Active objects */
} MAP;

/* Begin generated block */
__BEGIN_DECLS
extern DECLSPEC AG_ObjectClass mapClass;
extern DECLSPEC void MAP_InitSubsystem(void);
extern DECLSPEC void MAP_DestroySubsystem(void);
extern DECLSPEC MAP *MAP_New(void *, const char *);
extern DECLSPEC int MAP_AllocNodes(MAP *, Uint, Uint);
extern DECLSPEC void MAP_FreeNodes(MAP *);
extern DECLSPEC int MAP_Resize(MAP *, Uint, Uint);
extern DECLSPEC void MAP_SetZoom(MAP *, int, Uint);
extern DECLSPEC int MAP_PushLayer(MAP *, const char *);
extern DECLSPEC void MAP_PopLayer(MAP *);
extern DECLSPEC void MAP_InitLayer(MAP_Layer *, const char *);
extern DECLSPEC void MAP_InitCamera(MAP_Camera *, const char *);
extern DECLSPEC void MAP_InitModBlks(MAP *);
extern DECLSPEC int MAP_AddCamera(MAP *, const char *);
extern DECLSPEC void MAP_ModBegin(MAP *);
extern DECLSPEC void MAP_ModEnd(MAP *);
extern DECLSPEC void MAP_ModCancel(MAP *);
extern DECLSPEC void MAP_ModNodeChg(MAP *, int, int);
extern DECLSPEC void MAP_ModLayerAdd(MAP *, int);
extern DECLSPEC void MAP_Undo(MAP *);
extern DECLSPEC void MAP_Redo(MAP *);
extern DECLSPEC void MAP_ItemInit(MAP_Item *, enum map_item_type);
extern DECLSPEC void MAP_ItemSetCenter(MAP_Item *, int, int);
extern DECLSPEC void MAP_ItemSetMotion(MAP_Item *, int, int);
extern DECLSPEC void MAP_ItemSetFriction(MAP_Item *, int);
extern DECLSPEC void MAP_ItemSetLayer(MAP_Item *, int);
extern DECLSPEC void MAP_ItemDestroy(MAP *, MAP_Item *);
extern DECLSPEC int MAP_ItemLoad(MAP *, AG_DataSource *, MAP_Node *, MAP_Item **);
extern DECLSPEC void MAP_ItemSave(MAP *, AG_DataSource *, MAP_Item *);
extern DECLSPEC int MAP_ItemExtent(MAP *, MAP_Item *, AG_Rect *, int);
extern DECLSPEC void MAP_ItemDraw(MAP *, MAP_Item *, int, int, int);
extern DECLSPEC void MAP_ItemSetTile(MAP_Item *, MAP *, RG_Tileset *, Uint);
extern DECLSPEC void MAP_ItemSetAnim(MAP_Item *, MAP *, RG_Tileset *, Uint);
extern DECLSPEC void MAP_ItemAttrColor(Uint, int, AG_Color *);
extern DECLSPEC MAP_Item *MAP_ItemLocate(MAP *, int, int, int);
extern DECLSPEC void MAP_NodeInit(MAP_Node *);
extern DECLSPEC int MAP_NodeLoad(MAP *, AG_DataSource *, MAP_Node *);
extern DECLSPEC void MAP_NodeSave(MAP *, AG_DataSource *, MAP_Node *);
extern DECLSPEC void MAP_NodeDestroy(MAP *, MAP_Node *);
extern DECLSPEC void MAP_NodeRemoveAll(MAP *, MAP_Node *, int);
extern DECLSPEC void MAP_NodeCopy(MAP *, MAP_Node *, int, MAP *, MAP_Node *, int);
extern DECLSPEC void MAP_NodeMoveItem(MAP *, MAP_Node *, MAP_Item *, MAP *, MAP_Node *, int);
extern DECLSPEC MAP_Item *MAP_NodeCopyItem(const MAP_Item *, MAP *, MAP_Node *, int);
extern DECLSPEC void MAP_NodeMoveItemUp(MAP_Node *, MAP_Item *);
extern DECLSPEC void MAP_NodeMoveItemDown(MAP_Node *, MAP_Item *);
extern DECLSPEC void MAP_NodeMoveItemToTail(MAP_Node *, MAP_Item *);
extern DECLSPEC void MAP_NodeMoveItemToHead(MAP_Node *, MAP_Item *);
extern DECLSPEC void MAP_NodeDelItem(MAP *, MAP_Node *, MAP_Item *);
extern DECLSPEC void MAP_NodeSwapLayers(MAP *, MAP_Node *, int, int);
extern DECLSPEC MAP_Item *MAP_NodeAddTile(MAP *, MAP_Node *, RG_Tileset *, Uint32);
extern DECLSPEC MAP_Item *MAP_NodeAddAnim(MAP *, MAP_Node *, RG_Tileset *, Uint32);
extern DECLSPEC MAP_Item *MAP_NodeAddWarpPoint(MAP *, MAP_Node *, const char *, int, int, Uint8);
extern DECLSPEC void MAP_AttachActor(MAP *, struct map_actor *);
extern DECLSPEC void MAP_DetachActor(MAP *, struct map_actor *);
__END_DECLS
/* Close generated block */

#include <freesg/map/close.h>

#include <freesg/map/actor.h>
#include <freesg/map/tool.h>
#include <freesg/map/mapview.h>
#include <freesg/map/tools.h>
#include <freesg/map/mapedit.h>
#include <freesg/map/icons.h>
#include <freesg/map/map_math.h>

#endif /* _FREESG_MAP_MAP_H_ */
